@@ -17,12 +17,13 @@
 # $Id: client.py,v 1.52 2006/01/02 19:40:55 normanr Exp $
 
 """
-Client class establishs connection to XMPP Server and handles authentication
+Client class establishes connection to XMPP Server and handles authentication
 """
 
 import socket
-import transports_nb, dispatcher_nb, auth_nb, roster_nb, protocol, bosh
-from protocol import NS_TLS
+from . import transports_nb, dispatcher_nb, auth_nb, roster_nb, protocol, bosh
+from .protocol import NS_TLS
+from .auth_nb import SASL_AUTHENTICATION_MECHANISMS
 
 import logging
 log = logging.getLogger('nbxmpp.client_nb')
@@ -32,7 +33,7 @@ class NonBlockingClient:
     """
     Client class is XMPP connection mountpoint. Objects for authentication,
     network communication, roster, xml parsing ... are plugged to client object.
-    Client implements the abstract behavior - mostly negotioation and callbacks
+    Client implements the abstract behavior - mostly negotiation and callbacks
     handling, whereas underlying modules take care of feature-specific logic
     """
 
@@ -149,30 +150,33 @@ class NonBlockingClient:
 
     def connect(self, on_connect, on_connect_failure, hostname=None, port=5222,
     on_proxy_failure=None, on_stream_error_cb=None, proxy=None,
-    secure_tuple=('plain', None, None)):
+    secure_tuple=('tls', None, None, None)):
         """
         Open XMPP connection (open XML streams in both directions)
 
         :param on_connect: called after stream is successfully opened
-        :param on_connect_failure: called when error occures during connection
+        :param on_connect_failure: called when error occurs during connection
         :param hostname: hostname of XMPP server from SRV request
         :param port: port number of XMPP server
-        :param on_proxy_failure: called if error occurres during TCP connection to
-                proxy server or during proxy connecting process
+        :param on_proxy_failure: called if error occurs during TCP connection
+            to proxy server or during proxy connecting process
         :param proxy: dictionary with proxy data. It should contain at least
-                values for keys 'host' and 'port' - connection details for proxy serve
-                and optionally keys 'user' and 'pass' as proxy credentials
-        :param secure_tuple: tuple of (desired connection type, cacerts, mycerts)
-                connection type can be 'ssl' - TLS established after TCP connection,
-                'tls' - TLS established after negotiation with starttls, or 'plain'.
-                cacerts, mycerts - see tls_nb.NonBlockingTLS constructor for more
-                details
+            values for keys 'host' and 'port' - connection details for proxy
+            serve and optionally keys 'user' and 'pass' as proxy credentials
+        :param secure_tuple: tuple of (desired connection type, cacerts,
+            mycerts, tls_version, cipher_list)
+            connection type can be 'ssl' - TLS established after TCP connection,
+                'tls' - TLS established after negotiation with starttls, or
+                'plain'.
+            cacerts, mycerts, tls_version, cipher_list - see tls_nb.NonBlockingTLS
+                constructor for more details
         """
         self.on_connect = on_connect
         self.on_connect_failure=on_connect_failure
         self.on_proxy_failure = on_proxy_failure
         self.on_stream_error_cb = on_stream_error_cb
-        self.desired_security, self.cacerts, self.mycerts = secure_tuple
+        self.desired_security, self.cacerts, self.mycerts, self.tls_version, \
+            self.cipher_list = secure_tuple
         self.Connection = None
         self.Port = port
         self.proxy = proxy
@@ -209,6 +213,8 @@ class NonBlockingClient:
                         idlequeue=self.idlequeue,
                         estabilish_tls=establish_tls,
                         certs=certs,
+                        tls_version = self.tls_version,
+                        cipher_list = self.cipher_list,
                         proxy_creds=(proxy_user, proxy_pass),
                         xmpp_server=(self.xmpp_hostname, self.Port),
                         domain=self.Server,
@@ -230,6 +236,8 @@ class NonBlockingClient:
                     idlequeue=self.idlequeue,
                     estabilish_tls=establish_tls,
                     certs=certs,
+                    tls_version = self.tls_version,
+                    cipher_list = self.cipher_list,
                     proxy_dict=proxy_dict)
 
         # plug transport into client as self.Connection
@@ -249,13 +257,13 @@ class NonBlockingClient:
         try:
             self.ip_addresses = socket.getaddrinfo(hostname, port,
                     socket.AF_UNSPEC, socket.SOCK_STREAM)
-        except socket.gaierror, (errnum, errstr):
+        except socket.gaierror as exc:
             self.disconnect(message='Lookup failure for %s:%s, hostname: %s - %s' %
-                     (self.Server, self.Port, hostname, errstr))
-        except socket.error , (errnum, errstr):
-            # Catches an unexpected error with the socket
-            self.disconnect(message='General socket error for %s:%s, hostname: %s - %s' %
-                     (self.Server, self.Port, hostname, errstr))
+                     (self.Server, self.Port, hostname, str(exc)))
+        except socket.error as exc:
+            errnum, errstr = exc.errno, exc.strerror
+            self.disconnect(message='General socket error for %s:%s, hostname: '
+                '%s - %s' % (self.Server, self.Port, hostname, errstr))
         else:
             on_success()
 
@@ -324,7 +332,7 @@ class NonBlockingClient:
 
         if not mode:
             # starting state
-            if self.__dict__.has_key('Dispatcher'):
+            if 'Dispatcher' in self.__dict__:
                 self.Dispatcher.PlugOut()
                 self.got_features = False
             dispatcher_nb.Dispatcher.get_instance().PlugIn(self)
@@ -338,10 +346,9 @@ class NonBlockingClient:
                 self.Dispatcher.ProcessNonBlocking(data)
                 self.ip_addresses = []
             if not hasattr(self, 'Dispatcher') or \
-                    self.Dispatcher.Stream._document_attrs is None:
-                self._xmpp_connect_machine(
-                        mode='FAILURE',
-                        data='Error on stream open')
+            self.Dispatcher.Stream._document_attrs is None:
+                self._xmpp_connect_machine(mode='FAILURE',
+                    data='Error on stream open')
                 return
 
             # if terminating stanza was received after init request then client gets
@@ -419,7 +426,7 @@ class NonBlockingClient:
 
     def _tls_negotiation_handler(self, con=None, tag=None):
         """
-        Take care of TLS negotioation with <starttls>
+        Take care of TLS negotiation with <starttls>
         """
         log.info('-------------tls_negotiaton_handler() >> tag: %s' % tag)
         if not con and not tag:
@@ -443,7 +450,7 @@ class NonBlockingClient:
             # because <starttls> negotiation with BOSH is forbidden
             self.Connection.tls_init(
                     on_succ = lambda: self._xmpp_connect(socket_type='tls'),
-                    on_fail = lambda: self.disconnect('error while etabilishing TLS'))
+                    on_fail = lambda: self.disconnect('error while establishing TLS'))
 
     def _on_connect(self):
         """
@@ -465,7 +472,8 @@ class NonBlockingClient:
 ### follows code for authentication, resource bind, session and roster download
 ###############################################################################
 
-    def auth(self, user, password, resource='', sasl=True, on_auth=None):
+    def auth(self, user, password, resource='', sasl=True, on_auth=None, 
+             auth_mechs=None):
         """
         Authenticate connnection and bind resource. If resource is not provided
         random one or library name used
@@ -476,9 +484,27 @@ class NonBlockingClient:
         :param sasl: Boolean indicating if SASL shall be used. (default: True)
         :param on_auth: Callback, called after auth. On auth failure, argument
                 is None.
+        :param auth_mechs: Set of valid authentification mechanisms. If None all
+               authentification mechanisms will be allowed. Possible entries are:
+               'ANONYMOUS', 'EXTERNAL', 'GSSAPI', 'SCRAM-SHA-1-PLUS',
+               'SCRAM-SHA-1', 'DIGEST-MD5', 'PLAIN', 'X-MESSENGER-OAUTH2',
+               'XEP-0078'
         """
         self._User, self._Password = user, password
         self._Resource, self._sasl = resource, sasl
+        self._channel_binding = None
+        if self.connected in ('ssl', 'tls'):
+            try:
+                if self.protocol_type != 'BOSH':
+                    self._channel_binding = self.Connection.NonBlockingTLS.get_channel_binding()
+                    # TLS handshake is finished so channel binding data muss exist
+                    assert (self._channel_binding != None)
+            except NotImplementedError:
+                pass
+        if auth_mechs == None:
+            self._auth_mechs = SASL_AUTHENTICATION_MECHANISMS | set(['XEP-0078'])
+        else:
+            self._auth_mechs = auth_mechs
         self.on_auth = on_auth
         self._on_doc_attrs()
         return
@@ -510,10 +536,12 @@ class NonBlockingClient:
         """
         if self._sasl:
             auth_nb.SASL.get_instance(self._User, self._Password,
-                    self._on_start_sasl).PlugIn(self)
+                    self._on_start_sasl, self._channel_binding,
+                    self._auth_mechs).PlugIn(self)
         if not hasattr(self, 'SASL'):
             return
-        if not self._sasl or self.SASL.startsasl == 'not-supported':
+        if ('XEP-0078' in self._auth_mechs 
+            and (not self._sasl or self.SASL.startsasl == 'not-supported')):
             if not self._Resource:
                 self._Resource = 'xmpppy'
             auth_nb.NonBlockingNonSASL.get_instance(self._User, self._Password,
@@ -568,7 +596,7 @@ class NonBlockingClient:
         """
         Plug in the roster
         """
-        if not self.__dict__.has_key('NonBlockingRoster'):
+        if 'NonBlockingRoster' not in self.__dict__:
             return roster_nb.NonBlockingRoster.get_instance(version=version).PlugIn(self)
 
     def getRoster(self, on_ready=None, force=False):
@@ -576,7 +604,7 @@ class NonBlockingClient:
         Return the Roster instance, previously plugging it in and requesting
         roster from server if needed
         """
-        if self.__dict__.has_key('NonBlockingRoster'):
+        if 'NonBlockingRoster' in self.__dict__:
             return self.NonBlockingRoster.getRoster(on_ready, force)
         return None
 
